%{
/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2003, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/
package condor.classad;
import java.io.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
/** This is the internal classad parser generated by byacc.  From outside
 * this package, it should be accessed through the wrapper class ClassAdParser.
 * @see ClassAdParser
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.0
 */
class Parser {
%}
%token IDENTIFIER
%token INTEGER
%token REAL
%token STRING
%token TRUE
%token FALSE
%token ERROR
%token UNDEFINED

%left OP_LOR
%left OP_LAND
%left '|'
%left '^'
%left '&'
%left OP_EQ OP_NE OP_IS OP_ISNT
%left '<' '>' OP_LE OP_GE
%left OP_SHL OP_SHR OP_SHRR
%left '+' '-'
%left '*' '/' '%' 
%left UNARY

%type <Expr> Expr Expression PrefixExpr SuffixExpr Atom Literal
%type <FuncCall> Call
%type <RecordExpr> Record
%type <ListExpr> List
%type <List> ListBody DefinitionList
%type <AttrName> Identifier

%start Expression

%%
Expression
    : Expr 
    | Expr '?' Expression ':' Expression { $$ = new CondExpr($1, $3, $5); }
    ;
Expr
    : Expr OP_LOR Expr  { $$ = new Op(Expr.OR, $1, $3); }
    | Expr OP_LAND Expr  { $$ = new Op(Expr.AND, $1, $3); }
    | Expr '|' Expr  { $$ = new Op(Expr.BITOR, $1, $3); }
    | Expr '^' Expr  { $$ = new Op(Expr.BITXOR, $1, $3); }
    | Expr '&' Expr  { $$ = new Op(Expr.BITAND, $1, $3); }
    | Expr OP_EQ Expr  { $$ = new Op(Expr.EQUAL, $1, $3); }
    | Expr OP_NE Expr  { $$ = new Op(Expr.NOT_EQUAL, $1, $3); }
    | Expr OP_IS Expr  { $$ = new Op(Expr.SAME, $1, $3); }
    | Expr OP_ISNT Expr  { $$ = new Op(Expr.DIFFERENT, $1, $3); }
    | Expr '<' Expr  { $$ = new Op(Expr.LESS, $1, $3); }
    | Expr '>' Expr  { $$ = new Op(Expr.GREATER, $1, $3); }
    | Expr OP_LE Expr  { $$ = new Op(Expr.LESS_EQ, $1, $3); }
    | Expr OP_GE Expr  { $$ = new Op(Expr.GREATER_EQ, $1, $3); }
    | Expr OP_SHL Expr  { $$ = new Op(Expr.LEFT_SHIFT, $1, $3); }
    | Expr OP_SHR Expr  { $$ = new Op(Expr.RIGHT_SHIFT, $1, $3); }
    | Expr OP_SHRR Expr  { $$ = new Op(Expr.URIGHT_SHIFT, $1, $3); }
    | Expr '+' Expr  { $$ = new Op(Expr.PLUS, $1, $3); }
    | Expr '-' Expr  { $$ = new Op(Expr.MINUS, $1, $3); }
    | Expr '*' Expr { $$ = new Op(Expr.TIMES, $1, $3); }
    | Expr '/' Expr  { $$ = new Op(Expr.DIV, $1, $3); }
    | Expr '%' Expr  { $$ = new Op(Expr.MOD, $1, $3); }
    | PrefixExpr
    ;
PrefixExpr
    : SuffixExpr
    | '+' PrefixExpr %prec UNARY { $$ = new Op(Expr.UPLUS, $2); }
    | '-' PrefixExpr %prec UNARY { $$ = new Op(Expr.UMINUS, $2); }
    | '!' PrefixExpr %prec UNARY { $$ = new Op(Expr.NOT, $2); }
    | '~' PrefixExpr %prec UNARY { $$ = new Op(Expr.BIT_COMPLEMENT, $2); }
    ;
SuffixExpr
    : Atom
    | SuffixExpr '.' Identifier { $$ = new SelectExpr($1,  $3); }
    | SuffixExpr '[' Expression ']' { $$ = new SubscriptExpr($1, $3); }
    ;
Atom
    : Identifier { $$ = new AttrRef($1); }
    | Literal
    | List { $$ = $1; }
    | Record { $$ = $1; }
    | Call { $$ = $1; }
    | '(' Expression ')' { $$ = $2; }
    ;
List
    : '{' ListBody ListEnd { $$ = new ListExpr($2); }
    | '{' ListEnd { $$ = new ListExpr(null); }
    ;
ListEnd
    : '}'
    | ',' '}'
    ;
ListBody
    : Expression { List l = new ArrayList(); l.add($1); $$ = l; }
    | ListBody ',' Expression { List l = $1; l.add($3); $$ = l; }
    ;
Record
    : '[' DefinitionList RecordEnd {
        List l = $2;
        RecordExpr r = new RecordExpr(l.size());
        for (Iterator i = l.iterator(); i.hasNext(); ) {
            AttrName n = (AttrName) i.next();
            Expr e = (Expr) i.next();
            if (r.lookup(n) != null) {
                error("Duplicate definition of attribute " + n + " in Record");
            } else {
                r.insertAttribute(n, e);
            } /* if */
        } /* for */
        $$ = r;
    }
    | '[' RecordEnd { $$ = new RecordExpr(); }
    ;
RecordEnd
    : ']'
    | ';' ']'
    ;
DefinitionList
    : Identifier '=' Expression {
        List l = new ArrayList();
        l.add($1);
        l.add($3);
        $$ = l;
    }
    | DefinitionList ';' Identifier '=' Expression {
        List l = $1;
        l.add($3);
        l.add($5);
        $$ = l;
    }
    ;
Call
    : Identifier '(' ListBody ')' { $$ = new FuncCall($1, $3); }
    | Identifier '(' ')' { $$ = new FuncCall($1, null); }
    ;
Literal
    : REAL { $$ = lval; }
    | STRING { $$ = lval; }
    | INTEGER { $$ = lval; }
    | TRUE { $$ = Constant.True; }
    | FALSE { $$ = Constant.False; }
    | UNDEFINED { $$ = Constant.Undef; }
    | ERROR { $$ = Constant.Error; }
    ;
Identifier
    : IDENTIFIER { $$ = (AttrName) lval; }
    ;
%%
private static String VERSION = "$Id: classad.y,v 1.11 2003/01/20 21:05:04 solomon Exp $";

/** Verbosity for error messages.
 * Values are:
 * 0 - no messages
 * 1 - only show message and line number
 * 2 - also echo line in question
 * 3 - also list tokens expected
 */
/*package*/ int verbosity = 3;

/** Place to send error messages.  Default is System.err. */
PrintStream errs = System.err;

public Parser(Reader rdr) {
    reader = new LineNumberReader(rdr);
} // Parser(Reader)

public Parser(InputStream str) {
    reader = new LineNumberReader(new InputStreamReader(str));
} // Parser(InputStream)

public Parser(String str) {
    reader = new LineNumberReader(new StringReader(str));
} // Parser(String)

public void reset(Reader rdr) {
    reader = new LineNumberReader(rdr);
    lookahead = -1;
} // reset(Reader)

public void reset(InputStream str) {
    reader = new LineNumberReader(new InputStreamReader(str));
    lookahead = -1;
} // reset(InputStream)

public void reset(String str) {
    reader = new LineNumberReader(new StringReader(str));
    lookahead = -1;
} // reset(String)

// Lex support

/** Stream from which lex reads input.  */
private LineNumberReader reader;

/** The current input line. */
private String lineStr;

/** A version of lineStr unpacked into an array for speedy access */
private char[] line = new char[100];

/** The number of characters in the current input line. */
private int lineLen;

/** The current position in the current input line (0 = start of line) */
private int pos;

/** The position in the current line where the current token started. */
private int tokenStart;

/** The next input character to be consumed.  */
private int curChar;

/** A table of reserved words. */
private static final String[] reserved = {
    null,
    "error",
    "false",
    "is",
    "isnt",
    "true",
    "undefined",
};

/** A table of token codes parallel to the array reserved. */
private static final int[] reservedCode = {
    0, // unused
    ERROR,
    FALSE,
    OP_IS,
    OP_ISNT,
    TRUE,
    UNDEFINED
};

/** A trie constructed from the reserved words.
 * Three parallel arrays.  The element match[j] is supposed to match the
 * next letter of the input string.  If accept[j] = n != 0, then the string
 * matched thus far is reserved word n.  If the match[j] does not match
 * the next input character and alt[j] = n != 0, try match[n] instead.
 * If the input string starts with letter c, then start matching at position
 * trieStart[c], provided trieStart[c] > 0.
 */
private static char[] match;
private static int[] accept;
private static int[] alt;
private static int[] trieStart;
static {
    String m = " error false isnt true undefined ";
        //      012345678901234567890123456789012
        //                1         2         3
    match = m.toCharArray();

    trieStart = new int['z'];
    trieStart['e'] = trieStart['E'] = 1;
    trieStart['f'] = trieStart['F'] = 7;
    trieStart['i'] = trieStart['I'] = 13;
    trieStart['t'] = trieStart['T'] = 18;
    trieStart['u'] = trieStart['U'] = 23;

    accept = new int[m.length()];
    accept[6] = 1;      // error
    accept[12] = 2;     // false
    accept[15] = 3;     // is
    accept[17] = 4;     // isnt
    accept[22] = 5;     // true
    accept[32] = 6;     // undefined

    alt = new int[m.length()];
    alt[8] = 13;
};

/** A "wired-in" version of Character.toLowerCase(c) to avoid the
 * overhead in the common case of ASCII (c < 128).
 */
private static final int[] tolower = new int[] {
       0,   1,   2,   3,   4,   5,   6,   7,
       8,   9,  10,  11,  12,  13,  14,  15,
      16,  17,  18,  19,  20,  21,  22,  23,
      24,  25,  26,  27,  28,  29,  30,  31,
      32,  33,  34,  35,  36,  37,  38,  39,
      40,  41,  42,  43,  44,  45,  46,  47,
      48,  49,  50,  51,  52,  53,  54,  55,
      56,  57,  58,  59,  60,  61,  62,  63,
      64,  97,  98,  99, 100, 101, 102, 103,
     104, 105, 106, 107, 108, 109, 110, 111,
     112, 113, 114, 115, 116, 117, 118, 119,
     120, 121, 122,  91,  92,  93,  94,  95,
      96,  97,  98,  99, 100, 101, 102, 103,
     104, 105, 106, 107, 108, 109, 110, 111,
     112, 113, 114, 115, 116, 117, 118, 119,
     120, 121, 122, 123, 124, 125, 126, 127
};

/** Print a message indicated that some particular event occurred at a given
 * point in the input.
 * @param msg the message.
 */
private void lexdebug(Object msg) {
    System.out.println("LEX: line " + reader.getLineNumber()
        + " column " + (tokenStart+1)
        + " " + msg);
} // lexdebug(Object)

/** Get the "lookahead" token: the first token not yet consumed by the
 * parser.
 * The token is not removed from the input stream.
 * @return the code for the next token.
 * @see #getNextToken()
 */
public int nextToken() {
    if (lookahead < 0) {
        lookahead = lex();
    }
    return lookahead;
} // nextToken()

/** Get the value associated with the "lookahead" token: the first token not
 * yet consumed by the parser.
 * The token is not removed from the input stream.
 * @return the value associated with the next token.
 * @see #nextToken()
 */
public Object nextValue() {
    if (lookahead < 0) {
        lookahead = lex();
    }
    return lval;
} // nextValue()

/** Get and remove the "lookahead" token: the first token not yet consumed
 * by the parser.
 * The token is removed from the input stream.
 * @return the code for the next token.
 * @see #getNextToken()
 */
public int getNextToken() {
    if (lookahead < 0) {
        lookahead = lex();
    }
    int result = lookahead;
    lookahead = -1;
    return result;
} // getNextToken()

/** Gets the next line from reader and unpacks it to the array line[].
 * @return -1 for errors, 0 for EOF, or 1 for success.
 */
private final int getLine() {
    try {
        lineStr = reader.readLine();
    } catch (IOException e) {
        e.printStackTrace();
        return -1;
    }
    if (lineStr == null) {
        return 0; // EOF
    }

    lineLen = lineStr.length();
    
    if (line.length < lineLen + 1) {
        line = new char[lineLen + 20];
    }
    lineStr.getChars(0,lineLen,line,0);
    pos = 0;
    line[lineLen] = ' ';
    return 1;
} // getLine()

/** Appends a character to the current input line.  Used to recover from
 * unterminated strings.
 * @param c the character to append.
 */
private final void appendToInput(char c) {
    lineStr += "" + c;
    lineLen = lineStr.length();
    if (line.length < lineLen + 1) {
        line = new char[lineLen + 20];
    }
    lineStr.getChars(0,lineLen,line,0);
    line[lineLen] = ' ';
} // appendToInput(char)

/** Check whether the word stored in the line[] array starting at position
 * tokenStart matches one of the reserved words.  If so, return its index
 * in the array reserved[].  Otherwise, return 0.
 * @return an index in reserved[] or 0.
 */
private final int reservedLookup() {
    // Find the starting point in the trie
    int c = line[tokenStart];
    if (c >= trieStart.length) {
        return 0;
    }
    int pos = trieStart[c];
    if (pos++ == 0) {
        return 0;
    }
    // Start matching with the second character of the input
    for (int i = tokenStart+1; ;i++) {
        c = line[i];
        if (!Character.isJavaIdentifierStart((char)c)) {
            // End of input reached.  Return the code stored in the accept[]
            // array (will be zero if we are not at an "accept" node in the
            // trie.
            return accept[pos];
        }
        if (c >= tolower.length) {
            // Reserved words in the trie are entirely made up of ASCII
            // characters (<128).  If the input contains anything else, it is
            // obviously not a reserved word.
            return 0;
        }
        // Translate to lower case
        c = tolower[c];
        while (match[pos] != c) {
            // Match failed.  See if there is an alternative match at this
            // point (sibling node in the trie).
            pos = alt[pos];
            if (pos == 0) {
                // No alternative.  No match.
                return 0;
            }
        }
        // Successfully matched one more charcter of input.
        pos++;
    }
} // reservedLookup()

/** The lexical scanner.
 * It finds the next token in the input stream and returns its token code.
 * It also sets the fields tokenStart and pos so that the text of the token
 * is in line[tokenStart ... pos-1]. (For efficiency reasons, we do not
 * return the text as a String, since that would entail allocation a new String
 * that would, in most cases, immediately become garbage.)  
 * The "semantics" associated with the current token (if any) is stored
 * in lval.
 * <p>
 * @return the token code for the next token; 0 for EOF; -1 for error.
 */
private int lex() {
    int n;
    char quote;

    lval = null;
    // Skip spaces and comments
    for (;;) {
        if (pos == lineLen || (line[pos] == '/' && line[pos+1] == '/')) {
            if ((n = getLine()) <= 0) {
                return n;
            }
            continue;
        }
        if (line[pos] == '/' && line[pos+1] == '*') {
            for (pos += 2;;pos++) {
                if (pos >= lineLen) {
                    if ((n = getLine()) <= 0) {
                        return n;
                    }
                    pos = -1;
                    continue;
                }
                if (pos < lineLen - 1
                        && line[pos] == '*'
                        && line[pos+1] == '/')
                {
                    pos += 2;
                    break;
                }
            }
            continue;
        }
        if (!Character.isWhitespace(line[pos])) {
            break;
        }
        pos++;
    }
    tokenStart = pos;

    // Identifier or reserved word
    if (Character.isJavaIdentifierStart(line[pos])) {
        do {
            pos++;
        } while (Character.isJavaIdentifierPart(line[pos]));
        int i = reservedLookup();
        if (i > 0) {
            if (tracing) {
                lexdebug("reserved word " + reservedCode[i]
                         + " (" + reserved[i] + ")");
            }
            return reservedCode[i];
        }
        lval = AttrName.fromString(
                    new String(line, tokenStart, pos - tokenStart));
        if (tracing) {
            lexdebug("ID(" + lval + ")");
        }
        return IDENTIFIER;
    }

    // Number or '.'
    if (Character.isDigit(line[pos]) || line[pos]=='.') {
        // Number or lone '.'
        int type;
        if (line[pos] == '.' && !Character.isDigit(line[pos+1])) {
            pos++;
            if (tracing) {
                lexdebug("DOT");
            }
            return '.';
        }
        // Skip any leading digits
        for (; Character.isDigit(line[pos]); pos++) {
        }
        if (line[pos] == '.') {
            // There is a decimal point.  Skip the fraction part if any.
            type = REAL;
            for (pos++; Character.isDigit(line[pos]); pos++) {
            }
        } else {
            // There was no decimal point.  For now, assume it is an
            // integer.
            type = INTEGER;
            if (line[tokenStart] == '0'
                && (line[tokenStart+1] == 'x' || line[tokenStart+1] == 'x'))
            {
                // A hex integer constant.  Note that 'e' and 'E' are hex
                // digits, so 0x123e12 is parsed as a hex integer, and
                // and 0x123e+12 is parsed as three tokens: 0x123e + 12
                int value = 0;
                int digitValue;
                for (pos = tokenStart+2;
                    (digitValue = Character.digit(line[pos],16)) >= 0;
                    pos++)
                {
                    value = (value << 4) + digitValue;
                }
                // There must be at least one digit
                if (pos == tokenStart + 2) {
                    if (tracing) {
                        lexdebug("bad number");
                    }
                    return -1;
                }
                lval = Constant.getInstance(value);
                if (tracing) {
                    lexdebug("INTEGER " + lval);
                }
                return INTEGER;
            }
        }
        // Now look for an exponent
        if (line[pos] == 'e' || line[pos] == 'E') {
            type = REAL;
            pos++;
            if (line[pos] == '+' || line[pos] == '-') {
                pos++;
            }
            if (!Character.isDigit(line[pos])) {
                if (tracing) {
                    lexdebug("bad float");
                }
                return -1;
            }
            for (pos++; Character.isDigit(line[pos]); pos++) {
            }
        }
        // Finally, compute the value
        if (type == REAL) {
            lval = Constant.getInstance(
                        Double.parseDouble(
                            new String(line, tokenStart, pos - tokenStart)));
            if (tracing) {
                lexdebug("REAL " + lval);
            }
            return REAL;
        }
        // If there was no decimal point or exponent, we have to check
        // for a leading 0, meaning octal.
        int value = 0;
        int radix = (line[tokenStart] == '0') ? 8 : 10;
        for (int i=tokenStart; i<pos; i++) {
            int digitValue = Character.digit(line[i],radix);
            if (digitValue < 0) {
                if (tracing) {
                    lexdebug("bad oct");
                }
                return -1;
            }
            value = value * radix + digitValue;
        }
        lval = Constant.getInstance(value);
        if (tracing) {
            lexdebug("INT " + lval);
        }
        return INTEGER;
    } // Number

    // Multi-character operators
    switch (line[pos++]) {
    case '!':
        // != or !
        if (line[pos] == '=') {
            pos++;
            if (tracing) {
                lexdebug("NE");
            }
            return OP_NE;
        }
        break;
    case '&':
        // && or &
        if (line[pos] == '&') {
            pos++;
            if (tracing) {
                lexdebug("LAND");
            }
            return OP_LAND;
        }
        break;
    case '<':
        // << or <= or <
        if (line[pos] == '<') {
            pos++;
            if (tracing) {
                lexdebug("SHL");
            }
            return OP_SHL;
        }
        if (line[pos] == '=') {
            pos++;
            if (tracing) {
                lexdebug("LE");
            }
            return OP_LE;
        }
        break;
    case '=':
        // == or =
        if (line[pos] == '=') {
            pos++;
            if (tracing) {
                lexdebug("EQ");
            }
            return OP_EQ;
        }
        if (line[pos] == '?' && line[pos+1] == '=') {
            pos += 2;
            if (tracing) {
                lexdebug("IS");
            }
            return OP_IS;
        }
        if (line[pos] == '!' && line[pos+1] == '=') {
            pos += 2;
            if (tracing) {
                lexdebug("ISNT");
            }
            return OP_ISNT;
        }
        break;
    case '>':
        // >= or >> or >>> or >
        if (line[pos] == '=') {
            pos++;
            if (tracing) {
                lexdebug("GE");
            }
            return OP_GE;
        }
        if (line[pos] == '>') {
            pos++;
            if (line[pos] == '>') {
                pos++;
                if (tracing) {
                    lexdebug("SHRR");
                }
                return OP_SHRR;
            }
            if (tracing) {
                lexdebug("SHR");
            }
            return OP_SHR;
        }
        break;
    case '|':
        // || or |
        if (line[pos] == '|') {
            pos++;
            if (tracing) {
                lexdebug("LOR");
            }
            return OP_LOR;
        }
        break;
    case '"':
    case '\'':
        // String constant or quoted attribute name
        // Simply gather the characters up to an unescaped matching quote.
        // Other escape processing is done later.
        quote = line[pos - 1];
        for (;;) {
            if (pos == lineLen) {
                // Unterminated string constant.  Really should issue
                // a warning here!
                appendToInput((char) quote);
                pos++;
                break;
            }
            if (line[pos] == '\\') {
                pos +=2 ;
                continue;
            }
            if (line[pos] == quote) {
                pos++;
                break;
            }
            pos++;
        }
        if (quote == '"') {
            lval = Constant.getInstance(line, tokenStart + 1, pos - 1);
            if (tracing) {
                lexdebug("STRING(" + lval + ")");
            }
            return STRING;
        } else {
            lval = AttrName.fromText(
                                new String(line, tokenStart, pos - tokenStart));
            if (tracing) {
                lexdebug("ID(" + lval + ")");
            }
            return IDENTIFIER;
        }
    } // switch (line[pos++])
    // In all other cases, simply return the single character.  If it
    // isn't legit, the parser will catch it.
    if (tracing) {
        lexdebug("CHAR('" + line[pos-1] + "')");
    }
    return line[pos-1];
} // lex()

/** Print an error message, decorated with the current line and column number.
 * @param msg the message to print.
 */
private void error(String msg) {
    if (verbosity <= 0) {
        return;
    }
    if (verbosity == 1) {
        errs.println(msg
                + " at line " + reader.getLineNumber()
                + " column " + tokenStart);
        return;
    }
    errs.println(msg
            + " at line " + reader.getLineNumber()
            + " column " + tokenStart
            + ":");
    if (lineStr != null) {
        errs.println(lineStr);
        for (int i=0; i<tokenStart; i++) {
            errs.print(lineStr.charAt(i) == '\t' ? '\t' : ' ');
        }
        errs.println("^");
    }
    if (verbosity < 3) {
        return;
    }
    int[] valid = validTokens(state);
    if (valid.length > 0) {
        errs.print("Expecting ");
        for (int i=0; i<valid.length; i++) {
            if (i > 0) {
                errs.print(i == valid.length - 1 ? ", or " : ", ");
            }
            errs.print(tokenName(valid[i]));
        }
        errs.println(", but found " + tokenName(lookahead));
    } else {
        errs.println("Invalid token " + tokenName(lookahead));
    }
} // error(String)

/** Print a message, decorated with the current line and column number.
 * @param msg the message to print.
 */
/*package*/ void printMessage(String msg) {
    if (verbosity <= 0) {
        return;
    }
    errs.println(msg
                + " at line " + reader.getLineNumber()
                + " column " + tokenStart);
} // printMessage(String)
} // Parser
