#!/s/std/bin/perl -w
use strict 'subs';

# $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/genfunc.pl,v 1.3 2003/01/20 21:05:04 solomon Exp $

# Ad hoc script to generate tables of "builtin" functions.
#
# The table at the end of this file (after __END__) lists the set of
# function names and corresponding IDs for use as names of Java constants.
# Each line has the format
#   name ID comment
# A line in the input file containing __FUNC_NAMES__ is replaced by definition
# of a Java array of strings:
#   private static final String[] funcName = {
#       "name", // comment
#       "name", // comment
#       ...
#   }
# and a line containing __FUNC_ENUM__ is replaced by a set of Java constant
# definitions
#   private static final int NAME = 0;
#   private static final int NAME = 1;
#   ...
# such that funcName[X] = "x", where x and X are the function and corresponding
# ID from the table.  Each of these definitions is preceded by a javadoc
# version of the comment.
# /* ... */ comments and blank lines in the table are copied to the first
# array definition as is.

die "usage: $0 file.src\n"
    unless @ARGV == 1;
$fname = $ARGV[0];

die "$fname: $!\n" unless open(FI,"<$fname");

# pass 1: build tables from the stuff following __END__ below
$names = "";
$enums = "";
$n = 0;
while (<DATA>) {
    chomp();
    last if $_ eq "__END__";
    if (/\/\*/) {
        $names .= "\t\t$_\n";
        next;
    }
    if (/^$/) {
        $names .= "\n";
        next;
    }
    ($name,$id,$comment) = split(' ',$_,3);
    $names .= qq|\t\t"$name",\t// $comment\n|;
    $comment =~ s/&/&amp;/g;
    $comment =~ s/</&lt;/g;
    $comment =~ s/>/&gt;/g;
    $enums .= <<END;
    /** A constant representing the function $name.
     * $comment
     */
    private static final int $id = $n;
END
    $n++;
}
# pass 2, process input file, looking for keywords
print "/* Generated By:gengram.pl: Do not edit this file. */\n";
while (<FI>) {
    if (/__FUNC_NAMES__/) {
        $names =~ s/,([^,]*$)/$1/;  # remove last comma
        print "\tprivate static final String[] funcName = {\n$names\t};\n";
        next;
    }
    if (/__FUNC_ENUM__/) {
        print $enums;
        next;
    }
    print;
}
__END__
/* string operations */
strcat  STRCAT  varags; apply string() to each arg first
substr  SUBSTR  substr(stringval,offset,length)
strcmp  STRCMP  lex compare, return < = > 0
stricmp STRICMP like strcmp, but convert to lower case first
glob    GLOB    "glob" style matching with * and ?
iglob   IGLOB   like glob, but lower case in pattern matches either case

/* data conversion */
int     TOINT     convert arg to int (truc reals, parse strings)
real    TOREAL    convert arg to real (parse strings)
string  TOSTRING  convert arg to string

/* real to int (first apply real() if necessary) */
floor   FLOOR   greatest int <= value
ceiling CEILING smallest int >= value
round   ROUND   nearest int (IEEE rules)

/* time/date functions */
unixTime        TIME        no args; seconds since 1/1/1970, gmt
timeInterval    INTERVAL    convert to days+hh:mm:ss
localTimeString LCLTIME     Unix ctime format
gmtTimeString   GMTIME      Unix ctime format
relTime         RELTIME     convert from days+hh:mm:ss to relative time
absTime         ABSTIME     convert from ISO 8601 to absolute time
