// $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/tests/PrintTest.java,v 1.1 2003/01/20 19:47:23 solomon Exp $

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad.tests;
import condor.classad.*;
import java.util.*;
import java.io.*;

/** Simple test program to test parsing and printing classads.
 * Input consists of a sequence of classads separated by semicolons.
 */
class PrintTest {
    static public void p(Object o) { System.err.print(o); }
    static public void pl(Object o) { System.err.println(o); }

    /** Print a usage message and exit. */
    private void usage() {
        System.err.println("usage: java "
                        + getClass().getName() + " [-v [debug_level]] file");
        System.exit(1);
    } // usage()

    public static void main(String[] args) {
        int verbosity = 0; // amount of debugging output
        GetOpt opts = new GetOpt(getClass().getName(), args, "v::");
    opt_loop:
        for (;;) {
            switch (opts.nextOpt()) {
            default: usage();
            case 'v': // More verbose
                if (opts.optarg == null) {
                    verbosity++;
                } else {
                    verbosity = Integer.parseInt(opts.optarg);
                }
                break;
            }
        }

        if (opts.optind !=  args.length - 1) {
            usage();
        }

        Reader input = null;
        if (args.length > 0) {
            try {
                input = new FileReader(args[0]);
            } catch (Exception e) {
                e.printStackTrace();
                return;
            }
        } else {
            input = new InputStreamReader(System.in);
        }
        ClassAdParser parser = new ClassAdParser(input);
        //parser.enableTracing(true);
        ClassAdWriter out = new ClassAdWriter(System.out);

        // main loop
        for(;;) {
            Expr expr = parser.parse();
            if (expr == null) {
                break;
            }
            out.println(expr.toString());
            out.println(expr);
            out.flush();
            int lookahead = parser.getNextToken();
            if (lookahead != ';') {
                out.println("Expecting semicolon, found " + lookahead);
                return;
            }
        } // for (;;)
    } // main(String[])
} // PrintTest
